/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/**
 * \file   slist.h
 * \brief  Definition of a type-oblivious linear singly-linked list.
 * \author Sylvain Laperche <sylvain.laperche@gmail.com>
 * \date   2013/07/18 13:55:32
 */
#ifndef H_SLIST_LS_20130718135750
#define H_SLIST_LS_20130718135750

#include <stddef.h> /* For size_t, offsetof() and NULL. */

/** Head of a singly-linked list. */
typedef struct slist_s {
	size_t length;              /*!< List length. */
	struct slist_node_s* first; /*!< List head.   */
} slist;

/** A node of a singly-linked list. */
typedef struct slist_node_s {
	struct slist_node_s* next; /*!< Next element. */
} slist_node;

/**
 * Initializes the list.
 *
 * \param list a list.
 *
 * \pre `list` must be not NULL.
 *
 * \post `list` is empty.
 * \post `list` has a size of 0.
 *
 * \remarks Complexity: O(1)
 */
void slist_init(slist* list);

/**
 * Returns the size of the list.
 *
 * \param list a list.
 * \return the length of the list.
 *
 * \pre `list` must be not NULL.
 *
 * \remarks Complexity: O(1)
 */
size_t slist_length(const slist* list);

/**
 * Returns the first node.
 *
 * \param list a list.
 * \return the first node.
 *
 * \pre `list` must be not NULL.
 * \pre `list` must be not empty.
 *
 * \remarks Complexity: O(1)
 */
slist_node* slist_first(const slist* list);

/**
 * Returns the last node.
 *
 * \param list a list.
 * \return the last node.
 *
 * \pre `list` must be not NULL.
 *
 * \remarks Complexity: O(N)
 */
slist_node* slist_last(const slist* list);

/**
 * Returns the next node.
 *
 * \param node a node.
 * \return the next node.
 *
 * \pre `node` must be not NULL.
 *
 * \remarks Complexity: O(1)
 */
slist_node* slist_next(const slist_node* node);

/**
 * Returns the nth-node after the current one.
 *
 * \param node     a node.
 * \param distance distance to move on.
 * \return the nth-node after `node`.
 *
 * \pre `node` must be not NULL.
 *
 * \remarks Complexity: O(n)
 */
slist_node* slist_advance(slist_node* node, size_t distance);

/**
 * Returns node at the position `pos`.
 *
 * \param list a list.
 * \param pos  position (0-based) of the node.
 * \return the node at the position `pos`.
 *
 * \pre `list` must be not NULL.
 * \pre `list` must be not empty.
 * \pre `pos` must be in [0; slist_length(list)[.
 *
 * \remarks Complexity: O(N)
 */
slist_node* slist_at(const slist* list, size_t pos);

/**
 * Returns a pointer to the structure which contains the node.
 *
 * \param node      a list node (slist_node*).
 * \param type      type of the structure which contains the node.
 * \param fieldname name of the node (field name) in the structure.
 *
 * \pre `node` must be not NULL.
 *
 * \remarks Complexity: O(1)
 */
#define slist_elt(node, type, fieldname) \
	((type*)((char*)(node) - offsetof(type, fieldname)))

/**
 * Reverses the order of the elements in the list.
 *
 * \param list a list.
 *
 * \pre `list` must be not NULL.
 *
 * \remarks Complexity: O(N)
 */
void slist_reverse(slist* list);

/**
 * Swaps two lists.
 *
 * \param l1 list 1.
 * \param l2 list 2.
 *
 * \pre `l1` must be not NULL.
 * \pre `l2` must be not NULL.
 * \pre `l1` must be different of `l2`.
 *
 * \remarks Complexity: O(1)
 */
void slist_swap(slist* l1, slist* l2);

/**
 * Splits a list in two parts at the position `pos`.
 *
 * After the split:
 * \li `list` contains nodes in [0, pos[
 * \li `right` contains nodes in [pos,length(list)[
 *
 * Examples:
 * \code{.txt}
 * list = [1, 2, 3]
 * slist_split_at(list, 0, right) => ([],[1,2,3])
 * slist_split_at(list, 1, right) => ([1],[2,3])
 * slist_split_at(list, 2, right) => ([1,2],[3])
 * slist_split_at(list, 3, right) => ([1,2,3],[])
 * list = []
 * slist_split_at(list, 0, right) => ([],[])
 * \endcode
 *
 * \param list     list to split.
 * \param pos      position (0-based) where to split the list.
 * \param right an empty list to receive the part after `pos`
 *
 * \pre `list` must be not NULL.
 * \pre `right` must be not NULL.
 * \pre `right` must be empty.
 * \pre `list` must be different of `right`.
 *
 * \remarks Complexity: O(N)
 */
void slist_split_at(slist* list, size_t pos, slist* right);

/**
 * Concatenates two lists.
 *
 * \param l1 list 1.
 * \param l2 list 2.
 *
 * \pre `l1` must be not NULL.
 * \pre `l2` must be not NULL.
 * \pre `l1` must be different of `l2`.
 *
 * \post `l2` is reset to an empty list.
 *
 * \remarks Complexity: O(N)
 */
void slist_concat(slist* l1, slist* l2);

/**
 * Sort a list using a comparison function.
 *
 * The contents of the list are sorted in ascending order according to a
 * comparison function which is called with two arguments that point to the node
 * being compared.
 *
 * The comparison function must return an integer less than, equal to, or
 * greater than zero if the first argument is considered to be respectively less
 * than, equal to, or greater than the second.
 *
 * If two members compare as equal, their order in the sorted list is preserved.
 *
 * \param list list to sort.
 * \param cmp  comparison function to use.
 *
 * \pre `list` must be not NULL.
 * \pre `cmp` must be not NULL.
 *
 * \remarks The sort is stable. 
 * \remarks Complexity: O(N log N)
 * \remarks Space complexity: O(1)
 */
void slist_sort(slist* list, int (*cmp)(slist_node*, slist_node*));

/**
 * Iterates over list (nodes).
 * \param list a list (slist*).
 * \param curr a slist_node* used to hold the current element.
 *
 * \pre `list` must be not NULL.
 * \pre `curr` must be not NULL.
 *
 * \remarks Complexity: O(N)
 */
#define slist_foreach(list, curr) \
	for (curr = slist_first(list); curr != NULL; curr = slist_next(curr))

/**
 * Iterates over list (elements)
 * \param list      a list (slist*).
 * \param curr      pointer (type*) used to hold the current element.
 * \param type      type of the structure which contains the node.
 * \param fieldname name of the node (field name) in the structure.
 *
 * \pre `list` must be not NULL.
 * \pre `list` must be not empty.
 * \pre `curr` must be not NULL.
 *
 * \remarks Complexity: O(N)
 */
#define slist_foreach_elt(list, curr, type, fieldname)			   \
	for (curr = slist_elt(slist_first(list), type, fieldname);	   \
	     curr != NULL;						   \
	     curr = curr->fieldname.next == NULL ? NULL :		   \
	     slist_elt(slist_next(&(curr->fieldname)), type, fieldname))

/**
 * Iterates over list (nodes), allows deletion of the current node.
 * \param list a list (slist*).
 * \param curr a slist_node* used to hold the current element.
 * \param tmp  a slist_node* used as temporary storage.
 *
 * \pre `list` must be not NULL.
 * \pre `curr` must be not NULL.
 * \pre `tmp` must be not NULL.
 *
 * \remarks Complexity: O(N)
 */
#define slist_foreach_safe(list, curr, tmp)			\
	for (curr = slist_first(list), tmp = slist_next(curr);	\
	     curr != NULL;					\
	     curr = tmp, tmp = tmp == NULL ? NULL : slist_next(tmp))

/**
 * Iterates over list (elements), allows deletion of the current element.
 * \param list      a list (slist*).
 * \param curr      pointer (type*) used to hold the current element.
 * \param tmp       a slist_node* used as temporary storage.
 * \param type      type of the structure which contains the node.
 * \param fieldname name of the node (field name) in the structure.
 *
 * \pre `list` must be not NULL.
 * \pre `list` must be not empty.
 * \pre `curr` must be not NULL.
 *
 * \remarks Complexity: O(N)
 */
#define slist_foreach_elt_safe(list, curr, tmp, type, fieldname)	  \
	for (curr = slist_elt(slist_first(list), type, fieldname),	  \
	     tmp  = slist_next(&(curr->fieldname));			  \
	     curr != NULL;						  \
	     curr = tmp == NULL ? NULL : slist_elt(tmp, type, fieldname), \
	     tmp  = tmp == NULL ? NULL : slist_next(tmp))		  \

/**
 * Tests if a list is empty.
 *
 * \param list a list.
 * \return 1 if the list is empty, otherwise 0.
 *
 * \pre `list` must be not NULL.
 *
 * \remarks Complexity: O(1)
 */
unsigned char slist_is_empty(const slist* list);

/**
 * Adds `node` in the front of `list`
 *
 * \param list a list.
 * \param node the new node.
 *
 * \pre `list` must be not NULL.
 * \pre `node` must be not NULL.
 *
 * \post List size increases by 1.
 *
 * \remarks Complexity: O(1)
 */
void slist_add(slist* list, slist_node* node);

/**
 * Adds `node` at the end of `list`.
 *
 * \param list a list.
 * \param node the new node.
 *
 * \pre `list` must be not NULL.
 * \pre `node` must be not NULL.
 *
 * \post List size increases by 1.
 *
 * \remarks Complexity: O(n)
 */
void slist_append(slist* list, slist_node* node);

/**
 * Inserts `node` at after the node `curr`.
 *
 * \param list a list.
 * \param curr a node of `list`.
 * \param node new node.
 *
 * \pre `list` must be not NULL.
 * \pre `curr` must be not NULL.
 * \pre `curr` must a node of `list`.
 * \pre `node` must be not NULL.
 *
 * \post List size increases by 1.
 *
 * \remarks Complexity: O(1)
 */
void slist_insert_after(slist* list, slist_node* curr, slist_node* node);

/**
 * Inserts `node` at the position `pos` in `list`.
 *
 * \param list a list.
 * \param node new node.
 * \param pos  position (0-based) where to insert the new node.
 *
 * \pre `list` must be not NULL.
 * \pre `node` must be not NULL.
 * \pre `pos` must be in [0; slist_length(list)].
 *
 * \post List size increases by 1.
 *
 * \remarks Complexity: O(n)
 */
void slist_insert_at(slist* list, slist_node* node, size_t pos);

/**
 * Deletes the first node.
 *
 * \param list a list.
 *
 * \pre `list` must be not NULL.
 * \pre `list` must be not empty.
 *
 * \post List size decreases by 1.
 *
 * \remarks Complexity: O(1)
 */
void slist_del(slist* list);

/**
 * Deletes the node at after the node `curr`.
 *
 * \param list a list.
 * \param node a node of `list`.
 *
 * \pre `list` must be not NULL.
 * \pre `node` must be not NULL.
 * \pre `list` must be not empty.
 * \pre `node` must a node of `list`.
 *
 * \post List size decreases by 1.
 *
 * \remarks Complexity: O(1)
 */
void slist_del_after(slist* list, slist_node* node);

/**
 * Deletes the node at the position `pos`.
 *
 * \param list a list.
 * \param pos  index (0-based) of the node to delete.
 *
 * \pre `list` must be not NULL.
 * \pre `list` must be not empty.
 * \pre `pos` must be in [0; slist_length(list)[.
 *
 * \post List size decreases by 1.
 *
 * \remarks Complexity: O(n)
 */
void slist_del_at(slist* list, size_t pos);

#endif /* H_SLIST_LS_20130718135750 */
