/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/**
 * \file   slist.c
 * \brief  Implementation of a type-oblivious linear singly-linked list.
 * \author Sylvain Laperche <sylvain.laperche@gmail.com>
 * \date   2013/07/18 13:57:45
 */
#include <stddef.h> /* For NULL.   */
#include <string.h> /* For memcpy. */
#include <assert.h>

#include "slist.h"

void slist_init(slist* list)
{
	/* Check precondition. */
	assert(list != NULL);

	list->length = 0;
	list->first = NULL;
}

size_t slist_length(const slist* list)
{
	/* Check precondition. */
	assert(list != NULL);

	return list->length;
}

slist_node* slist_first(const slist* list)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(list->length > 0);

	return list->first;
}

slist_node* slist_last(const slist* list)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(list->length > 0);

	return slist_at(list, list->length - 1);
}

slist_node* slist_next(const slist_node* node)
{
	/* Check precondition. */
	assert(node != NULL);

	return node->next;
}

slist_node* slist_advance(slist_node* node, size_t distance)
{
	/* Check precondition. */
	assert(node != NULL);

	{
		size_t i = 0;
		while (i < distance && node != NULL) {
			node = slist_next(node);
			++i;
		}

		return node;
	}
}

slist_node* slist_at(const slist* list, size_t pos)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(list->length > 0);
	assert(pos < list->length);

	{
		slist_node* curr = list->first;
		size_t i = 0;

		while (i < pos) {
			curr = curr->next;
			++i;
		}
		return curr;
	}
}

unsigned char slist_is_empty(const slist* list)
{
	/* Check precondition. */
	assert(list != NULL);

	return list->length == 0;
}

void slist_reverse(slist* list)
{
	/* Check precondition. */
	assert(list != NULL);

	{
		slist_node* prev = NULL;
		slist_node* next = NULL;
		slist_node* curr = list->first;

		while (curr != NULL) {
			next = curr->next;
			curr->next = prev;
			prev = curr;
			curr = next;
		}
		list->first = prev;
	}
}

void slist_swap(slist* l1, slist* l2)
{
	/* Check preconditions. */
	assert(l1 != NULL);
	assert(l2 != NULL);
	assert(l1 != l2);

	{
		slist tmp;
		memcpy(&tmp, l1, sizeof *l1);
		memcpy(l1,   l2, sizeof *l2);
		memcpy(l2, &tmp, sizeof tmp);
	}
}

void slist_split_at(slist* list, size_t pos, slist* right)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(right != NULL);
	assert(right->length == 0);
	assert(list != right);

	/* `pos` is capped at length(list). */
	if (pos > slist_length(list))
		pos = slist_length(list);
	/*
	 * Special case: split at 0.
	 * Initial state: (list, right) = ([A B C D], [])
	 *   Final state: (list, right) = ([], [A B C D])
	 * => swap the list.
	 */
	if (pos == 0)
		slist_swap(list, right);
	/*
	 * Special case: split at the end.
	 * Initial state: (list, right) = ([A B C D], [])
	 *   Final state: (list, right) = ([A B C D], [])
	 * => nothing to do.
	 */
	else if (pos == list->length)
		;
	else {
		slist_node* cutting_point = slist_at(list, pos-1);

		right->first = cutting_point->next;
		right->length = list->length - pos;

		cutting_point->next = NULL;
		list->length = pos;
	}
}

void slist_concat(slist* l1, slist* l2)
{
	/* Check preconditions. */
	assert(l1 != NULL);
	assert(l2 != NULL);
	assert(l1 != l2);

	/*
	 * Special case: concat to an empty list.
	 * Initial state: (l1, l2) = ([], [A B C D])
	 *   Final state: (l1, l2) = ([A B C D], [])
	 * => swap the list.
	 */
	if (slist_is_empty(l1))
		slist_swap(l1, l2);
	else {
		slist_node* last = slist_last(l1);

		last->next = l2->first;
		l2->first = NULL;

		l1->length += l2->length;
		l2->length = 0;
	}

}

void slist_sort(slist* list, int (*cmp)(slist_node*, slist_node*))
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(cmp  != NULL);
	/* A bottom-up implementation of the merge sort. */
	{
		slist_node* tail;  /* Tail of the merged sorted list. */
		slist_node* p;     /* Beginning of the first run.     */
		slist_node* p_end; /* End of the first run.           */
		slist_node* q;     /* Beginning of the second run.    */
		slist_node* q_end; /* End of the second run.          */
		size_t run_size;   /* Size of the runs to process.    */
		size_t length = list->length;
		/* For each size of run in 1, 2, 4, ... length/2 */
		for (run_size = 1; run_size < length; run_size *= 2) {
			p = list->first;
			tail = NULL;
			/* Iterate on each pair of runs (p and q) of size `run_size`*/
			while (p != NULL) {
				slist_node* node; /* Node to add into the merged run. */

				q = slist_advance(p, run_size);
				p_end = q;
				/* If length(list) is odd, `q` may be empty: handle this. */
				q_end = q == NULL ? NULL : slist_advance(q, run_size);
				/*Merging the two runs. */
				while (p != p_end || q != q_end) {
					/*
					 * We select for insertion the node from the first run if:
					 * - the second run is over
					 * - the first run is not over AND
					 *   current node of 1st run <= current node of 2nd run.
					 * Using '<=' instead of '<' ensure sorting stability.
					 */
					if (q == q_end || (p != p_end && cmp(p, q) <= 0)) {
						node = p;
						p = p->next;
					}
					else {
						node = q;
						q = q->next;
					}
					/* Append the new node to the merged sorted list. */
					if (tail != NULL)
						tail->next = node;
					else /* First insertion: update the head. */
						list->first = node;
					tail = node; /* Update the tail pointer. */
				}
				p = q_end; /* The next run starts after the current pair. */
				tail->next = NULL; /* Finalization of the merged list.    */
			}
		}
	}
}

void slist_add(slist* list, slist_node* node)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(node != NULL);

	node->next = list->first;
	list->first = node;
	list->length++;
}

void slist_append(slist* list, slist_node* node)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(node != NULL);

	if (slist_is_empty(list))
		slist_add(list, node);
	else
		slist_insert_after(list, slist_last(list), node);
}

void slist_insert_after(slist* list, slist_node* curr, slist_node* node)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(curr != NULL);
	assert(node != NULL);

	node->next = curr->next;
	curr->next = node;
	list->length++;
}

void slist_insert_at(slist* list, slist_node* node, size_t pos)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(node != NULL);
	assert(pos <= list->length);

	/* Special case: insert in front. */
	if (pos == 0)
		slist_add(list, node);
	else {
		slist_node* curr = slist_at(list, pos - 1);
		slist_insert_after(list, curr, node);
	}
}

void slist_del(slist* list)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(list->length > 0);

	list->first = list->first->next;
	list->length--;
}

void slist_del_after(slist* list, slist_node* node)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(node != NULL);
	assert(list->length > 0);

	node->next = node->next->next;
	list->length--;
}

void slist_del_at(slist* list, size_t pos)
{
	/* Check preconditions. */
	assert(list != NULL);
	assert(list->length > 0);
	assert(pos < list->length);

	/* Special case: delete in front. */
	if (pos == 0)
		slist_del(list);
	else {
		slist_node* curr = slist_at(list, pos - 1);
		slist_del_after(list, curr);
	}
}
