/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/*
 *    Filename: access_tests.c
 * Description: Implements test cases for "access" functions.
 *     Created: 2013/08/20 13:09:33
 *      Author: Sylvain Laperche <sylvain.laperche@gmail.com>
 */
#include <stddef.h> /* For NULL. */
#include <signal.h> /* For SIGABRT. */

#include "access_tests.h"

/******************
 * SETUP/TEARDOWN *
 ******************/

/* List used for the test. */
static slist list;

/* Initialize the list. */
static void setup (void)
{
	slist_init(&list);
}

/* Nothing to do. */
static void teardown (void)
{
}



/***************
 * SLIST_FIRST *
 ***************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_first_null)
{
	slist_node* node = slist_first(NULL);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Violate the precondition "the list must be not empty." */
START_TEST(test_first_empty)
{
	slist_node* node = slist_first(&list);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Normal use. */
START_TEST(test_first)
{
	int_node* first;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Insert in front. */
	slist_add(&list, &node1.node);
	first = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(first->i, 1);
	/* Append: first does not change. */
	slist_append(&list, &node2.node);
	first = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(first->i, 1);
	/* Insert in front: first change. */
	slist_add(&list, &node3.node);
	first = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(first->i, 3);
}
END_TEST



/**************
 * SLIST_LAST *
 **************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_last_null)
{
	slist_node* node = slist_last(NULL);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Violate the precondition "the list must be not empty." */
START_TEST(test_last_empty)
{
	slist_node* node;
	node = slist_last(&list);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Normal use. */
START_TEST(test_last)
{
	int_node* last;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Insert in front. */
	slist_add(&list, &node1.node);
	last = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(last->i, 1);
	/* Append: last change. */
	slist_append(&list, &node2.node);
	last = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(last->i, 2);
	/* Insert in front: last does not change. */
	slist_add(&list, &node3.node);
	last = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(last->i, 2);
}
END_TEST



/**************
 * SLIST_NEXT *
 **************/

/* Violate the precondition "the node must be not NULL." */
START_TEST(test_next_null)
{
	slist_node* node = slist_next(NULL);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Normal use. */
START_TEST(test_next)
{
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Populate the list. */
	slist_append(&list, &node1.node);
	slist_append(&list, &node2.node);
	slist_append(&list, &node3.node);
	/* Check the node. */
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_next(&(node->node)), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_next(&(node->node)), int_node, node);
	ck_assert_int_eq(node->i, 3);
}
END_TEST



/*****************
 * SLIST_ADVANCE *
 *****************/

/* Violate the precondition "the node must be not NULL." */
START_TEST(test_advance_null)
{
	slist_node* node = slist_advance(NULL, 1);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Capping. */
START_TEST(test_advance_capping)
{
	slist_node* curr;
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	/* Populate the list. */
	slist_append(&list, &node1.node);
	slist_append(&list, &node2.node);
	/* Advance more than the list length: returns NULL. */
	curr = slist_first(&list);
	curr = slist_advance(curr, 42);
	ck_assert(curr == NULL);
}
END_TEST

/* Normal use. */
START_TEST(test_advance)
{
	slist_node* curr;
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Populate the list. */
	slist_append(&list, &node1.node);
	slist_append(&list, &node2.node);
	slist_append(&list, &node3.node);
	/* Advance from the first node. */
	curr = slist_first(&list);
	node = slist_elt(curr, int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_advance(curr, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_advance(curr, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* Advance from the second node. */
	curr = slist_at(&list, 1);
	node = slist_elt(slist_advance(curr, 1), int_node, node);
	ck_assert_int_eq(node->i, 3);
}
END_TEST



/************
 * SLIST_AT *
 ************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_at_null)
{
	slist_node* node = slist_at(NULL, 0);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Violate the precondition "the list must be not empty." */
START_TEST(test_at_empty)
{
	slist_node* node;
	node = slist_at(&list, 0);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Violate the precondition "index must be in [0; slist_length(list)[" */
START_TEST(test_at_bad_index)
{
	slist_node* node;
	int_node node1 = { 1, { NULL } };

	slist_append(&list, &node1.node);

	node = slist_at(&list, 1);
	(void)node; /* Avoid a warning. */
}
END_TEST

/* Normal use. */
START_TEST(test_at)
{
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Populate the list. */
	slist_append(&list, &node1.node);
	slist_append(&list, &node2.node);
	slist_append(&list, &node3.node);
	/* Check the value of each node. */
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
}
END_TEST



/**************
 * TEST CASES *
 **************/
TCase* access_test_cases(void)
{
	TCase* tc_access = tcase_create("access");

	tcase_add_checked_fixture(tc_access, setup, teardown);

	tcase_add_test_raise_signal(tc_access, test_first_null , SIGABRT);
	tcase_add_test_raise_signal(tc_access, test_first_empty, SIGABRT);
	tcase_add_test(tc_access, test_first);

	tcase_add_test_raise_signal(tc_access, test_last_null , SIGABRT);
	tcase_add_test_raise_signal(tc_access, test_last_empty, SIGABRT);
	tcase_add_test(tc_access, test_last);

	tcase_add_test_raise_signal(tc_access, test_next_null , SIGABRT);
	tcase_add_test(tc_access, test_next);

	tcase_add_test_raise_signal(tc_access, test_advance_null , SIGABRT);
	tcase_add_test(tc_access, test_advance_capping);
	tcase_add_test(tc_access, test_advance);

	tcase_add_test_raise_signal(tc_access, test_at_null ,     SIGABRT);
	tcase_add_test_raise_signal(tc_access, test_at_empty,     SIGABRT);
	tcase_add_test_raise_signal(tc_access, test_at_bad_index, SIGABRT);
	tcase_add_test(tc_access, test_at);

	return tc_access;
}
