/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/*
 *    Filename: add_tests.c
 * Description: Implements test cases for "add" functions.
 *     Created: 2013/08/20 13:03:26
 *      Author: Sylvain Laperche <sylvain.laperche@gmail.com>
 */
#include <stddef.h> /* For NULL. */
#include <signal.h> /* For SIGABRT. */

#include "add_tests.h"

/******************
 * SETUP/TEARDOWN *
 ******************/

/* List used for the test. */
static slist list;

/* Initialize the list. */
static void setup (void)
{
	slist_init(&list);
}

/* Nothing to do. */
static void teardown (void)
{
}


/*************
 * SLIST_ADD *
 *************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_add_list_null)
{
	int_node node;
	slist_add(NULL, &node.node);
}
END_TEST

/* Violate the precondition "the node to add must be not NULL." */
START_TEST(test_add_node_null)
{
	slist_add(&list, NULL);
}
END_TEST

/* Normal use. */
START_TEST(test_add)
{
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	/* Add a node. first and last should be equal. */
	slist_add(&list, &node1.node);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 1);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	/* Insert in front: first change, not last */
	slist_add(&list, &node2.node);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 2);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
}
END_TEST



/****************
 * SLIST_APPEND *
 ****************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_append_list_null)
{
	int_node node;
	slist_append(NULL, &node.node);
}
END_TEST

/* Violate the precondition "the node to add must be not NULL." */
START_TEST(test_append_node_null)
{
	slist_append(&list, NULL);
}
END_TEST

/* Normal use. */
START_TEST(test_append)
{
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	/* Add a node. first and last should be equal. */
	slist_append(&list, &node1.node);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 1);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	/* Append: last change, not first */
	slist_append(&list, &node2.node);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 2);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(node->i, 2);
}
END_TEST



/**********************
 * SLIST_INSERT_AFTER *
 **********************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_insert_after_list_null)
{
	int_node node;
	slist_node curr;

	slist_insert_after(NULL, &curr, &node.node);
}
END_TEST

/* Violate the precondition "the current node must be not NULL." */
START_TEST(test_insert_after_curr_null)
{
	int_node node;
	slist_insert_after(&list, NULL, &node.node);
}
END_TEST

/* Violate the precondition "the node to insert must be not NULL." */
START_TEST(test_insert_after_node_null)
{
	slist_node curr;
	slist_insert_after(&list, &curr, NULL);
}
END_TEST

/* Normal use. */
START_TEST(test_insert_after)
{
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Insert the first node. */
	slist_add(&list, &node1.node);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 1);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	/* Insert a node after the first node. */
	slist_insert_after(&list, &node1.node, &node2.node);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 2);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	/* Insert a node after the second node. */
	slist_insert_after(&list, &node2.node, &node3.node);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
}
END_TEST



/*******************
 * SLIST_INSERT_AT *
 *******************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_insert_at_list_null)
{
	slist_node node;
	slist_insert_at(NULL, &node, 0);
}
END_TEST

/* Violate the precondition "the node to insert must be not NULL." */
START_TEST(test_insert_at_node_null)
{
	slist_insert_at(&list, NULL, 0);
}
END_TEST

/* Violate the precondition "index must be in [0; slist_length(list)]" */
START_TEST(test_insert_at_bad_index)
{
	int_node node1 = { 1, { NULL } };
	slist_insert_at(&list, &node1.node, 1);
}
END_TEST

/* Normal use. */
START_TEST(test_insert_at)
{
	int_node* node;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Insert a node at the position 0. */
	slist_insert_at(&list, &node2.node, 0);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 1);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 2);
	/* Insert a node at the position 1. */
	slist_insert_at(&list, &node3.node, 1);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 2);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* Insert a node at the position 2. */
	slist_insert_at(&list, &node1.node, 2);
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 1);
}
END_TEST



/**************
 * TEST CASES *
 **************/
TCase* add_test_cases(void)
{
	TCase* tc_add = tcase_create("add");

	tcase_add_checked_fixture(tc_add, setup, teardown);

	tcase_add_test_raise_signal(tc_add, test_add_list_null, SIGABRT);
	tcase_add_test_raise_signal(tc_add, test_add_node_null, SIGABRT);
	tcase_add_test(tc_add, test_add);

	tcase_add_test_raise_signal(tc_add, test_append_list_null, SIGABRT);
	tcase_add_test_raise_signal(tc_add, test_append_node_null, SIGABRT);
	tcase_add_test(tc_add, test_append);

	tcase_add_test_raise_signal(tc_add, test_insert_after_list_null, SIGABRT);
	tcase_add_test_raise_signal(tc_add, test_insert_after_curr_null, SIGABRT);
	tcase_add_test_raise_signal(tc_add, test_insert_after_node_null, SIGABRT);
	tcase_add_test(tc_add, test_insert_after);


	tcase_add_test_raise_signal(tc_add, test_insert_at_list_null, SIGABRT);
	tcase_add_test_raise_signal(tc_add, test_insert_at_node_null, SIGABRT);
	tcase_add_test_raise_signal(tc_add, test_insert_at_bad_index, SIGABRT);
	tcase_add_test(tc_add, test_insert_at);

	return tc_add;
}
