/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/*
 *    Filename: core_tests.c
 * Description: Implements test cases for core functions.
 *     Created: 2013/08/20 12:53:18
 *      Author: Sylvain Laperche <sylvain.laperche@gmail.com>
 */
#include <stddef.h> /* For NULL.    */
#include <signal.h> /* For SIGABRT. */

#include "core_tests.h"

/**************
 * SLIST_INIT *
 **************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_init_null)
{
	slist_init(NULL);
}
END_TEST

/* Normal use. */
START_TEST(test_init)
{
	slist list;
	slist_init(&list);
	/* Check that the list is empty. */
	ck_assert_int_eq(slist_length(&list), 0);
	ck_assert(slist_is_empty(&list));
}
END_TEST



/****************
 * SLIST_LENGTH *
 ****************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_length_null)
{
	slist_length(NULL);
}
END_TEST

/* Normal use. */
START_TEST(test_length)
{
	slist list;
	int_node node1 = { 1, { NULL } };
	int_node node2 = { 2, { NULL } };
	int_node node3 = { 3, { NULL } };
	/* Empty. */
	slist_init(&list);
	ck_assert_int_eq(slist_length(&list), 0);
	/* 1 node. */
	slist_add(&list, &node1.node);
	ck_assert_int_eq(slist_length(&list), 1);
	/* 2 node. */
	slist_add(&list, &node2.node);
	ck_assert_int_eq(slist_length(&list), 2);
	/* 3 node. */
	slist_add(&list, &node3.node);
	ck_assert_int_eq(slist_length(&list), 3);
	/* 2 node. */
	slist_del(&list);
	ck_assert_int_eq(slist_length(&list), 2);
	/* 1 node. */
	slist_del(&list);
	ck_assert_int_eq(slist_length(&list), 1);
	/* Empty. */
	slist_del(&list);
	ck_assert_int_eq(slist_length(&list), 0);
}
END_TEST



/***************
 * SLIST_ENTRY *
 ***************/

/* Normal use. */
START_TEST(test_elt)
{
	int_node node = { 42, { NULL } };
	int_node* node_ptr = slist_elt(&node.node, int_node, node);
	/* Check value. */
	ck_assert_int_eq(node_ptr->i, 42);
}
END_TEST



/******************
 * SLIST_IS_EMPTY *
 ******************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_is_empty_null)
{
	int empty = slist_is_empty(NULL);
	(void)empty;
}
END_TEST

/* Normal use. */
START_TEST(test_is_empty)
{
	slist list;
	int_node node;
	/* Empty list. */
	slist_init(&list);
	ck_assert(slist_is_empty(&list));
	/* Non-empty list. */
	slist_add(&list, &node.node);
	ck_assert(!slist_is_empty(&list));
}
END_TEST



/**************
 * TEST CASES *
 **************/
TCase* core_test_cases(void)
{
	TCase* tc_core = tcase_create("core");

	tcase_add_test_raise_signal(tc_core, test_init_null, SIGABRT);
	tcase_add_test(tc_core, test_init);

	tcase_add_test_raise_signal(tc_core, test_length_null, SIGABRT);
	tcase_add_test(tc_core, test_length);

	tcase_add_test(tc_core, test_elt);

	tcase_add_test_raise_signal(tc_core, test_is_empty_null, SIGABRT);
	tcase_add_test(tc_core, test_is_empty);

	return tc_core;
}
