/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/*
 *    Filename: del_tests.c
 * Description: Implements test cases for "del" functions.
 *     Created: 2013/08/20 13:14:58
 *      Author: Sylvain Laperche <sylvain.laperche@gmail.com>
 */
#include <stddef.h> /* For NULL. */
#include <signal.h> /* For SIGABRT. */

#include "del_tests.h"

/******************
 * SETUP/TEARDOWN *
 ******************/

/* List used for the test. */
static slist list;
/* List nodes. */
static int_node node1 = { 1, { NULL } };
static int_node node2 = { 2, { NULL } };
static int_node node3 = { 3, { NULL } };

/* Populate the list. */
static void setup (void)
{
	slist_init(&list);
	slist_add(&list, &node3.node);
	slist_add(&list, &node2.node);
	slist_add(&list, &node1.node);
}

/* Nothing to do. */
static void teardown (void)
{
}



/*************
 * SLIST_DEL *
 *************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_del_null)
{
	slist_del(NULL);
}
END_TEST

/* Violate the precondition "the list must be not empty." */
START_TEST(test_del_empty)
{
	slist_init(&list);
	slist_del(&list);
}
END_TEST

/* Normal use. */
START_TEST(test_del)
{
	int_node* node;
	/* The list is complete: 3 nodes => [1, 2, 3]. */
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 1st node => [2, 3]. */
	slist_del(&list);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 2);
	ck_assert_int_eq(slist_length(&list), 2);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 1st node => [3]. */
	slist_del(&list);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 3);
	ck_assert_int_eq(slist_length(&list), 1);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 1st node => []. */
	slist_del(&list);
	ck_assert_int_eq(slist_length(&list), 0);
	ck_assert(slist_is_empty(&list));
}
END_TEST



/*******************
 * SLIST_DEL_AFTER *
 *******************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_del_after_list_null)
{
	slist_node node;
	slist_del_after(NULL, &node);
}
END_TEST

/* Violate the precondition "the current node must be not NULL." */
START_TEST(test_del_after_node_null)
{
	slist_del_after(&list, NULL);
}
END_TEST

/* Violate the precondition "the list must be not empty." */
START_TEST(test_del_after_empty)
{
	slist_node node;

	slist_init(&list);
	slist_del_after(&list, &node);
}
END_TEST

/* Normal usage. */
START_TEST(test_del_after)
{
	int_node* node;
	/* The list is complete: 3 nodes => [1, 2, 3]. */
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 3rd node => [1, 2]. */
	slist_del_after(&list, &node2.node);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(node->i, 2);
	ck_assert_int_eq(slist_length(&list), 2);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 2nd node => [1]. */
	slist_del_after(&list, &node1.node);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	ck_assert_int_eq(slist_length(&list), 1);
	ck_assert(!slist_is_empty(&list));
	/* Delete the first node => []. */
	slist_del(&list);
	ck_assert_int_eq(slist_length(&list), 0);
	ck_assert(slist_is_empty(&list));
}
END_TEST



/****************
 * SLIST_DEL_AT *
 ****************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_del_at_null)
{
	slist_del_at(NULL, 0);
}
END_TEST

/* Violate the precondition "the list must be not empty." */
START_TEST(test_del_at_empty)
{
	slist_init(&list);
	slist_del_at(&list, 0);
}
END_TEST

/* Violate the precondition "index must be in [0; slist_length(list)[" */
START_TEST(test_del_at_bad_index)
{
	slist_del_at(&list, 3);
}
END_TEST

/* Normal use. */
START_TEST(test_del_at)
{
	int_node* node;
	/* The list is complete: 3 nodes => [1, 2, 3]. */
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 2nd node => [1, 3]. */
	slist_del_at(&list, 1);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_last(&list), int_node, node);
	ck_assert_int_eq(node->i, 3);
	ck_assert_int_eq(slist_length(&list), 2);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 2nd (last) node => [1]. */
	slist_del_at(&list, 1);
	node = slist_elt(slist_first(&list), int_node, node);
	ck_assert_int_eq(node->i, 1);
	ck_assert_int_eq(slist_length(&list), 1);
	ck_assert(!slist_is_empty(&list));
	/* Delete the 1st node => []. */
	slist_del_at(&list, 0);
	ck_assert_int_eq(slist_length(&list), 0);
	ck_assert(slist_is_empty(&list));
}
END_TEST



/**************
 * TEST CASES *
 **************/
TCase* del_test_cases(void)
{
	TCase* tc_del = tcase_create("del");

	tcase_add_checked_fixture(tc_del, setup, teardown);

	tcase_add_test_raise_signal(tc_del, test_del_null , SIGABRT);
	tcase_add_test_raise_signal(tc_del, test_del_empty, SIGABRT);
	tcase_add_test(tc_del, test_del);

	tcase_add_test_raise_signal(tc_del, test_del_after_list_null, SIGABRT);
	tcase_add_test_raise_signal(tc_del, test_del_after_node_null, SIGABRT);
	tcase_add_test_raise_signal(tc_del, test_del_after_empty    , SIGABRT);
	tcase_add_test(tc_del, test_del_after);

	tcase_add_test_raise_signal(tc_del, test_del_at_null     , SIGABRT);
	tcase_add_test_raise_signal(tc_del, test_del_at_empty    , SIGABRT);
	tcase_add_test_raise_signal(tc_del, test_del_at_bad_index, SIGABRT);
	tcase_add_test(tc_del, test_del_at);

	return tc_del;
}
