/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/*
 *    Filename: foreach_tests.c
 * Description: Implements test cases for "foreach" functions.
 *     Created: 2013/08/20 13:16:32
 *      Author: Sylvain Laperche <sylvain.laperche@gmail.com>
 */
#include <stddef.h> /* For NULL and size_t. */

#include "foreach_tests.h"

/******************
 * SETUP/TEARDOWN *
 ******************/

/* List used for the test. */
static slist list;
/* List nodes. */
static int_node node1 = { 1, { NULL } };
static int_node node2 = { 2, { NULL } };
static int_node node3 = { 3, { NULL } };
/* Reference content. */
static const int ref[] = { 1, 2, 3 };

/* Populate the list. */
static void setup (void)
{
	slist_init(&list);
	slist_add(&list, &node3.node);
	slist_add(&list, &node2.node);
	slist_add(&list, &node1.node);
}

/* Nothing to do. */
static void teardown (void)
{
}



/*****************
 * SLIST_FOREACH *
 *****************/

/* Normal use. */
START_TEST(test_foreach)
{
	size_t i = 0;
	slist_node* curr;
	/* Traverse the list. */
	slist_foreach(&list, curr)
	{
		int_node* elt = slist_elt(curr, int_node, node);
		ck_assert_int_eq(elt->i, ref[i]);
		++i;
	}
	/* Check that we traverse all the elements. */
	ck_assert_int_eq(i, slist_length(&list));
}
END_TEST



/*********************
 * SLIST_FOREACH_ELT *
 *********************/

/* Normal use. */
START_TEST(test_foreach_elt)
{
	size_t i = 0;
	int_node* curr;
	/* Traverse the list. */
	slist_foreach_elt(&list, curr, int_node, node)
	{
		ck_assert_int_eq(curr->i, ref[i]);
		++i;
	}
	/* Check that we traverse all the elements. */
	ck_assert_int_eq(i, slist_length(&list));
}
END_TEST



/**********************
 * SLIST_FOREACH_SAFE *
 **********************/

/* Normal use. */
START_TEST(test_foreach_safe)
{
	size_t i = 0;
	slist_node* curr;
	slist_node* tmp;
	/* Traverse the list. */
	slist_foreach_safe(&list, curr, tmp)
	{
		int_node* elt = slist_elt(curr, int_node, node);
		ck_assert_int_eq(elt->i, ref[i]);
		++i;
	}
	/* Check that we traverse all the elements. */
	ck_assert_int_eq(i, slist_length(&list));
}
END_TEST

/* Clear a list. */
START_TEST(test_foreach_safe_del)
{
	size_t i = 0;
	slist_node* curr;
	slist_node* tmp;

	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(!slist_is_empty(&list));
	/* Clear the list. */
	slist_foreach_safe(&list, curr, tmp)
	{
		int_node* elt = slist_elt(curr, int_node, node);
		ck_assert_int_eq(elt->i, ref[i]);
		slist_del(&list);
		++i;
	}
	/* Check that we delete all the elements. */
	ck_assert_int_eq(slist_length(&list), 0);
	ck_assert(slist_is_empty(&list));
}
END_TEST



/**************************
 * SLIST_FOREACH_ELT_SAFE *
 **************************/

/* Normal use. */
START_TEST(test_foreach_elt_safe)
{
	size_t i = 0;
	slist_node* tmp;
	int_node* elt;
	/* Traverse the list. */
	slist_foreach_elt_safe(&list, elt, tmp, int_node, node)
	{
		ck_assert_int_eq(elt->i, ref[i]);
		++i;
	}
	/* Check that we traverse all the elements. */
	ck_assert_int_eq(i, slist_length(&list));
}
END_TEST

/* Clear a list. */
START_TEST(test_foreach_elt_safe_del)
{
	size_t i = 0;
	slist_node* tmp;
	int_node* elt;

	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(!slist_is_empty(&list));
	/* Clear the list. */
	slist_foreach_elt_safe(&list, elt, tmp, int_node, node)
	{
		ck_assert_int_eq(elt->i, ref[i]);
		slist_del(&list);
		++i;
	}
	/* Check that we delete all the elements. */
	ck_assert_int_eq(slist_length(&list), 0);
	ck_assert(slist_is_empty(&list));
}
END_TEST



/**************
 * TEST CASES *
 **************/
TCase* foreach_test_cases(void)
{
	TCase* tc_foreach = tcase_create("foreach");

	tcase_add_checked_fixture(tc_foreach, setup, teardown);

	tcase_add_test(tc_foreach, test_foreach);

	tcase_add_test(tc_foreach, test_foreach_elt);

	tcase_add_test(tc_foreach, test_foreach_safe);
	tcase_add_test(tc_foreach, test_foreach_safe_del);

	tcase_add_test(tc_foreach, test_foreach_elt_safe);
	tcase_add_test(tc_foreach, test_foreach_elt_safe_del);

	return tc_foreach;
}
