/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/*
 *    Filename: operation_tests.c
 * Description: Implements test cases for the operation on the list.
 *     Created: 2013/08/23 19:43:48
 *      Author: Sylvain Laperche <sylvain.laperche@gmail.com>
 */
#include <stddef.h> /* For NULL and size_t. */
#include <signal.h> /* For SIGABRT.         */

#include "operation_tests.h"

/******************
 * SETUP/TEARDOWN *
 ******************/

/* Lists used for the test. */
static slist list;
static slist right;

/* List nodes. */
static int_node node1 = { 1, { NULL } };
static int_node node2 = { 2, { NULL } };
static int_node node3 = { 3, { NULL } };

static int cmp_by_value(slist_node* n1, slist_node* n2)
{
	int2_node* p1 = slist_elt(n1, int2_node, node);
	int2_node* p2 = slist_elt(n2, int2_node, node);

	return p1->value - p2->value;
}

/* Populate the list. */
static void setup (void)
{
	/* Initialize lists. */
	slist_init(&list);
	slist_init(&right);
	/* Populate the list. */
	slist_add(&list, &node3.node);
	slist_add(&list, &node2.node);
	slist_add(&list, &node1.node);
}

/* Nothing to do. */
static void teardown (void)
{
}



/*****************
 * SLIST_REVERSE *
 *****************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_reverse_null)
{
	slist_reverse(NULL);
}
END_TEST

/* Normal use. */
START_TEST(test_reverse)
{
	int_node* node;
	/* `list` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* Reverse. */
	slist_reverse(&list);
	/* `list` = [3, 2, 1]. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 3);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 1);
	/* Empty list. */
	slist_reverse(&right);
}
END_TEST




/**************
 * SLIST_SWAP *
 **************/

/* Violate the precondition "the list 1 must be not NULL." */
START_TEST(test_swap_l1_null)
{
	slist_swap(NULL, &list);
}
END_TEST

/* Violate the precondition "the list 2 must be not NULL." */
START_TEST(test_swap_l2_null)
{
	slist_swap(&list, NULL);
}
END_TEST

/* Violate the precondition "list 1 != list 2." */
START_TEST(test_swap_same)
{
	slist_swap(&list, &list);
}
END_TEST

/* Normal use. */
START_TEST(test_swap)
{
	int_node* node;
	/* `list` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* `right` = []. */
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
	/* Swap. */
	slist_swap(&list, &right);
	/* `list` = []. */
	ck_assert(slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 0);
	/* `right` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 3);
	node = slist_elt(slist_at(&right, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&right, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&right, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);

}
END_TEST



/******************
 * SLIST_SPLIT_AT *
 ******************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_split_at_list_null)
{
	slist_split_at(NULL, 0, &right);
}
END_TEST

/* Violate the precondition "the right list must be not NULL." */
START_TEST(test_split_at_right_null)
{
	slist_split_at(&list, 0, NULL);
}
END_TEST

/* Violate the precondition "the right list must be empty." */
START_TEST(test_split_at_not_empty)
{
	/* Add a node into `right`. */
	slist_add(&right, &node1.node);
	slist_split_at(&list, 0, &right);
}
END_TEST

/* Violate the precondition "list 1 != list 2." */
START_TEST(test_split_at_same)
{
	slist_split_at(&list, 0, &list);
}
END_TEST

/* Empty list. */
START_TEST(test_split_empty)
{
	/* Initialize the list. */
	slist_init(&list);
	/* Initial state. */
	ck_assert(slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 0);
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
	/* slist_split_at(list, 0, right) => ([],[]) */
	slist_split_at(&list, 0, &right);
	/* Check `list` empty. */
	ck_assert(slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 0);
	/* Check `right` empty. */
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
}
END_TEST

/* Split before the first node. */
START_TEST(test_split_before_first)
{
	int_node* node;
	/* Initial state. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
	/* slist_split_at(list, 0, right) => ([],[1,2,3]) */
	slist_split_at(&list, 0, &right);
	/* Check `list` empty. */
	ck_assert(slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 0);
	/* Check `right` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 3);
	node = slist_elt(slist_at(&right, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&right, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&right, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
}
END_TEST

/* Split after the last node. */
START_TEST(test_split_after_last)
{
	int_node* node;
	/* Initial state. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
	/* slist_split_at(list, 3, right) => ([1,2,3],[]) */
	slist_split_at(&list, 3, &right);
	/* Check `list` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* Check `right` empty. */
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
}
END_TEST

/* Split after the last element: capping. */
START_TEST(test_split_capping)
{
	int_node* node;
	/* Initial state. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
	/* slist_split_at(list, 3, right) => ([1,2,3],[]) */
	slist_split_at(&list, 42, &right);
	/* Check `list` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* Check `right` empty. */
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
}
END_TEST


/* Normal use. */
START_TEST(test_split_at)
{
	int_node* node;
	/* Initial state. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
	/* slist_split_at(list, 2, right) => ([1,2],[3]) */
	slist_split_at(&list, 2, &right);
	/* Check `list` = [1, 2]. */
	ck_assert(!slist_is_empty(&list));
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	/* Check `right` = [3]. */
	ck_assert_int_eq(slist_length(&list), 2);
	ck_assert(!slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 1);
	node = slist_elt(slist_at(&right, 0), int_node, node);
	ck_assert_int_eq(node->i, 3);
}
END_TEST



/****************
 * SLIST_CONCAT *
 ****************/

/* Violate the precondition "the list 1 must be not NULL." */
START_TEST(test_concat_l1_null)
{
	slist_concat(NULL, &list);
}
END_TEST

/* Violate the precondition "the list 2 must be not NULL." */
START_TEST(test_concat_l2_null)
{
	slist_concat(&list, NULL);
}
END_TEST

/* Violate the precondition "list 1 != list 2." */
START_TEST(test_concat_same)
{
	slist_concat(&list, &list);
}
END_TEST

/* Concatention with empty list. */
START_TEST(test_concat_empty)
{
	slist empty1;
	slist empty2;
	int_node* node;
	/* Initialize lists. */
	slist_init(&empty1);
	slist_init(&empty2);

	/* [] + [] = []. */
	slist_concat(&empty1, &empty2);
	/* Check `empty1` = []. */
	ck_assert(slist_is_empty(&empty1));
	ck_assert_int_eq(slist_length(&empty1), 0);
	/* Check `empty2` = []. */
	ck_assert(slist_is_empty(&empty2));
	ck_assert_int_eq(slist_length(&empty2), 0);

	/* [1, 2, 3] + [] = [1, 2, 3]. */
	slist_concat(&list, &empty1);
	/* Check `list` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 3);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* Check `empty1` = []. */
	ck_assert(slist_is_empty(&empty1));
	ck_assert_int_eq(slist_length(&empty1), 0);

	/* [] + [1, 2, 3] = [1, 2, 3]. */
	slist_concat(&empty1, &list);
	/* Check `empty1` = [1, 2, 3]. */
	ck_assert(!slist_is_empty(&empty1));
	ck_assert_int_eq(slist_length(&empty1), 3);
	node = slist_elt(slist_at(&empty1, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&empty1, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&empty1, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	/* Check `list` = []. */
	ck_assert(slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 0);
}
END_TEST

/* Normal use. */
START_TEST(test_concat)
{
	int_node* node;
	int_node node4 = { 4, { NULL } };
	int_node node5 = { 5, { NULL } };
	int_node node6 = { 6, { NULL } };

	slist_add(&right, &node6.node);
	slist_add(&right, &node5.node);
	slist_add(&right, &node4.node);
	/* [1, 2, 3] + [4, 5, 6] = [1, 2, 3, 4, 5, 6]. */
	slist_concat(&list, &right);
	/* Check `list` = [1, 2, 3, 4, 5, 6]. */
	ck_assert(!slist_is_empty(&list));
	ck_assert_int_eq(slist_length(&list), 6);
	node = slist_elt(slist_at(&list, 0), int_node, node);
	ck_assert_int_eq(node->i, 1);
	node = slist_elt(slist_at(&list, 1), int_node, node);
	ck_assert_int_eq(node->i, 2);
	node = slist_elt(slist_at(&list, 2), int_node, node);
	ck_assert_int_eq(node->i, 3);
	node = slist_elt(slist_at(&list, 3), int_node, node);
	ck_assert_int_eq(node->i, 4);
	node = slist_elt(slist_at(&list, 4), int_node, node);
	ck_assert_int_eq(node->i, 5);
	node = slist_elt(slist_at(&list, 5), int_node, node);
	ck_assert_int_eq(node->i, 6);
	/* Check `right` = []. */
	ck_assert(slist_is_empty(&right));
	ck_assert_int_eq(slist_length(&right), 0);
}
END_TEST



/**************
 * SLIST_SORT *
 **************/

/* Violate the precondition "the list must be not NULL." */
START_TEST(test_sort_list_null)
{
	slist_sort(NULL, cmp_by_value);
}
END_TEST

/* Violate the precondition "the comparison function must be not NULL." */
START_TEST(test_sort_cmp_null)
{
	slist_sort(&list, NULL);
}
END_TEST

/* Empty list. */
START_TEST(test_sort_empty)
{
	slist int2_list;
	slist_init(&int2_list);
	/* Sort an empty list. */
	slist_sort(&int2_list, cmp_by_value);
}
END_TEST

/* List with a single element. */
START_TEST(test_sort_single)
{
	slist int2_list;
	int2_node node = { 42, 1, { NULL } };
	int2_node* p;
	/* Populate the list. */
	slist_init(&int2_list);
	slist_add(&int2_list, &node.node);
	/* Initial state. */
	p = slist_elt(slist_at(&int2_list, 0), int2_node, node);
	ck_assert_int_eq(p->value, 42);
	ck_assert_int_eq(p->position, 1);
	/* Sort a list with only 1 node. */
	slist_sort(&int2_list, cmp_by_value);
	/* Final state (should be equal to the initial state. */
	p = slist_elt(slist_at(&int2_list, 0), int2_node, node);
	ck_assert_int_eq(p->value, 42);
	ck_assert_int_eq(p->position, 1);
}
END_TEST

/* Normal use (even number of elements). */
START_TEST(test_sort_even)
{
	int2_node* node;
	int2_node data[] = { { 7, 1, { NULL } }, { 5, 1, { NULL } },
			     { 0, 1, { NULL } }, { 3, 1, { NULL } },
			     { 3, 2, { NULL } }, { 0, 2, { NULL } },
			     { 6, 1, { NULL } }, { 9, 1, { NULL } } };
	int result[][2] = { { 0, 1 }, { 0, 2 }, { 3, 1 }, { 3, 2 },
			    { 5, 1 }, { 6, 1 }, { 7, 1 }, { 9, 1 } };

	size_t tabsize = sizeof data / sizeof data[0];
	size_t i;
	slist int2_list;
	/* Populate the list. */
	slist_init(&int2_list);
	for (i = 0; i < tabsize; ++i)
		slist_append(&int2_list, &data[i].node);
	/* Initial state. */
	ck_assert(!slist_is_empty(&int2_list));
	ck_assert_int_eq(slist_length(&int2_list), tabsize);
	for (i = 0; i < tabsize; ++i)
	{
		node = slist_elt(slist_at(&int2_list, i), int2_node, node);
		ck_assert_int_eq(node->value,    data[i].value);
		ck_assert_int_eq(node->position, data[i].position);
	}
	/* Sort the list. */
	slist_sort(&int2_list, cmp_by_value);
	/* Final state. */
	ck_assert(!slist_is_empty(&int2_list));
	ck_assert_int_eq(slist_length(&int2_list), tabsize);
	for (i = 0; i < tabsize; ++i)
	{
		node = slist_elt(slist_at(&int2_list, i), int2_node, node);
		ck_assert_int_eq(node->value,    result[i][0]);
		ck_assert_int_eq(node->position, result[i][1]);
	}
}
END_TEST

/* Normal use (odd number of elements). */
START_TEST(test_sort_odd)
{
	int2_node* node;
	int2_node data[] = { { 7, 1, { NULL } }, { 5, 1, { NULL } },
			     { 0, 1, { NULL } }, { 3, 1, { NULL } },
			     { 3, 2, { NULL } }, { 0, 2, { NULL } },
			     { 6, 1, { NULL } }, { 9, 1, { NULL } },
			     { 7, 2, { NULL } } };
	int result[][2] = { { 0, 1 }, { 0, 2 }, { 3, 1 }, { 3, 2 },
			    { 5, 1 }, { 6, 1 }, { 7, 1 }, { 7, 2 }, { 9, 1 } };

	size_t tabsize = sizeof data / sizeof data[0];
	size_t i;
	slist int2_list;
	/* Populate the list. */
	slist_init(&int2_list);
	for (i = 0; i < tabsize; ++i)
		slist_append(&int2_list, &data[i].node);
	/* Initial state. */
	ck_assert(!slist_is_empty(&int2_list));
	ck_assert_int_eq(slist_length(&int2_list), tabsize);
	for (i = 0; i < tabsize; ++i)
	{
		node = slist_elt(slist_at(&int2_list, i), int2_node, node);
		ck_assert_int_eq(node->value,    data[i].value);
		ck_assert_int_eq(node->position, data[i].position);
	}
	/* Sort the list. */
	slist_sort(&int2_list, cmp_by_value);
	/* Final state. */
	ck_assert(!slist_is_empty(&int2_list));
	ck_assert_int_eq(slist_length(&int2_list), tabsize);
	for (i = 0; i < tabsize; ++i)
	{
		node = slist_elt(slist_at(&int2_list, i), int2_node, node);
		ck_assert_int_eq(node->value,    result[i][0]);
		ck_assert_int_eq(node->position, result[i][1]);
	}
}
END_TEST

/* Normal use (already sorted). */
START_TEST(test_sort_sorted)
{
	int2_node* node;
	int2_node data[] = { { 0, 1, { NULL } }, { 0, 2, { NULL } },
			     { 3, 1, { NULL } }, { 3, 2, { NULL } },
			     { 5, 1, { NULL } }, { 6, 1, { NULL } },
			     { 7, 1, { NULL } }, { 7, 2, { NULL } },
			     { 9, 1, { NULL } } };
	size_t tabsize = sizeof data / sizeof data[0];
	size_t i;
	slist int2_list;
	/* Populate the list. */
	slist_init(&int2_list);
	for (i = 0; i < tabsize; ++i)
		slist_append(&int2_list, &data[i].node);
	/* Initial state. */
	ck_assert(!slist_is_empty(&int2_list));
	ck_assert_int_eq(slist_length(&int2_list), tabsize);
	for (i = 0; i < tabsize; ++i)
	{
		node = slist_elt(slist_at(&int2_list, i), int2_node, node);
		ck_assert_int_eq(node->value,    data[i].value);
		ck_assert_int_eq(node->position, data[i].position);
	}
	/* Sort the list. */
	slist_sort(&int2_list, cmp_by_value);
	/* Final state. */
	ck_assert(!slist_is_empty(&int2_list));
	ck_assert_int_eq(slist_length(&int2_list), tabsize);
	for (i = 0; i < tabsize; ++i)
	{
		node = slist_elt(slist_at(&int2_list, i), int2_node, node);
		ck_assert_int_eq(node->value,    data[i].value);
		ck_assert_int_eq(node->position, data[i].position);
	}
}
END_TEST



/**************
 * TEST CASES *
 **************/
TCase* operation_test_cases(void)
{
	TCase* tc_op = tcase_create("operation");

	tcase_add_checked_fixture(tc_op, setup, teardown);

	tcase_add_test_raise_signal(tc_op, test_reverse_null, SIGABRT);
	tcase_add_test(tc_op, test_reverse);

	tcase_add_test_raise_signal(tc_op, test_swap_l1_null, SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_swap_l2_null, SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_swap_same   , SIGABRT);
	tcase_add_test(tc_op, test_swap);

	tcase_add_test_raise_signal(tc_op, test_split_at_list_null,  SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_split_at_right_null, SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_split_at_not_empty,  SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_split_at_same     ,  SIGABRT);
	tcase_add_test(tc_op, test_split_empty);
	tcase_add_test(tc_op, test_split_before_first);
	tcase_add_test(tc_op, test_split_after_last);
	tcase_add_test(tc_op, test_split_capping);
	tcase_add_test(tc_op, test_split_at);

	tcase_add_test_raise_signal(tc_op, test_concat_l1_null, SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_concat_l2_null, SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_concat_same   , SIGABRT);
	tcase_add_test(tc_op, test_concat_empty);
	tcase_add_test(tc_op, test_concat);

	tcase_add_test_raise_signal(tc_op, test_sort_list_null, SIGABRT);
	tcase_add_test_raise_signal(tc_op, test_sort_cmp_null,  SIGABRT);
	tcase_add_test(tc_op, test_sort_empty);
	tcase_add_test(tc_op, test_sort_single);
	tcase_add_test(tc_op, test_sort_even);
	tcase_add_test(tc_op, test_sort_odd);
	tcase_add_test(tc_op, test_sort_sorted);

	return tc_op;
}
