/***********************************************************************
 * Copyright (c) 2013, Sylvain LAPERCHE                                *
 * All rights reserved.                                                *
 * License: BSD 3-Clause (http://opensource.org/licenses/BSD-3-Clause) *
 ***********************************************************************/
/*
 *    Filename: run_tests.c
 * Description: Runs the unit test suite.
 *     Created: 2013/08/19 15:03:45
 *      Author: Sylvain Laperche <sylvain.laperche@gmail.com>
 */
#include <stdlib.h> /* For EXIT_SUCCESS and EXIT_FAILURE. */
#include <stdio.h>  /* For stderr and fprintf().          */
#include <errno.h>  /* For errno and strerror().          */

#include <sys/resource.h> /* For setrlimit(). */

#include <check.h>

#include "core_tests.h"
#include "access_tests.h"
#include "foreach_tests.h"
#include "operation_tests.h"
#include "add_tests.h"
#include "del_tests.h"

/*
 * The following constants come from:
 * http://sourceforge.net/p/predef/wiki/OperatingSystems/
 */
/* Windows. */
#if defined(_WIN32) || defined(_WIN64)
#define REDIRECT_STDERR redirect_stderr("NUL")
#define DISABLE_COREDUMP ((void) 0)
/* Linux and BSDs. */
#elif defined(__linux__)   || defined(__FreeBSD__) || defined(__NetBSD__) || \
      defined(__OpenBSD__) || defined(__APPLE__)   || defined(__DragonFly__)
#define REDIRECT_STDERR redirect_stderr("/dev/null")
#define DISABLE_COREDUMP disable_coredump()
/* Other systems. */
#else
#define REDIRECT_STDERR fputs("Warning: cannot redirect stderr.\n", stderr)
#define DISABLE_COREDUMP ((void) 0)
#endif

/*
 * Setup the test suite.
 *
 * \return the test suite.
 */
static Suite* test_suite(void)
{
	Suite* suite = suite_create("slist");

	suite_add_tcase(suite, core_test_cases());
	suite_add_tcase(suite, access_test_cases());
	suite_add_tcase(suite, foreach_test_cases());
	suite_add_tcase(suite, operation_test_cases());
	suite_add_tcase(suite, add_test_cases());
	suite_add_tcase(suite, del_test_cases());

	return suite;
}

/*
 * Disable the core file generation for the current process and its children.
 */
static void disable_coredump(void)
{
	struct rlimit core_limit = { 0, 0 };

	if (setrlimit(RLIMIT_CORE, &core_limit) == -1)
		fprintf(stderr, "setrlimit: %s\n"
				"Warning: Cannot disable coredump\n",
			strerror(errno));
}

/*
 * Redirect stderr to the null file.
 */
static void redirect_stderr(const char* nullfile)
{
	if (freopen(nullfile, "w", stderr) == NULL)
		fprintf(stderr, "freopen: %s\n"
				"Warning: Cannot redirect stderr\n",
			strerror(errno));
}

/*
 * Runs the unit test suite.
 *
 * \return EXIT_SUCCESS if all the tests ran well, otherwise EXIT_FAILURE.
 */
int main(void)
{
	int nb_failure = 0;
	Suite* suite = test_suite();
	SRunner* runner = srunner_create(suite);

	/*
	 * The library I test here uses assert() to check some preconditions (if
	 * they are violated, assert() kills the programm by calling abort()).
	 *
	 * My test suite checks the behavior when these preconditions are
	 * violated, thus it may kill the program several times and (because of
	 * abort()) generates a coredump each time.
	 * This is quite annoying, so I disable the coredump generation.
	 *
	 * I also redirect stderr in order to avoid visual pollution by failed
	 * assertions during the tests (Check writes all its output on stdout,
	 * so the redirection of stderr is not an issue)
	 */
	DISABLE_COREDUMP;
	REDIRECT_STDERR;

	srunner_run_all(runner, CK_NORMAL);
	nb_failure = srunner_ntests_failed(runner);

	srunner_free(runner);

	return (nb_failure == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
